import sys
from random import shuffle
import argparse

import numpy as np
import scipy.io

from keras.models import Sequential, Graph
from keras.layers.core import Dense, Dropout, Activation, Reshape
from keras.optimizers import SGD
from keras.layers.recurrent import LSTM
from keras.utils import np_utils, generic_utils
from keras.preprocessing import sequence

from sklearn import preprocessing
from sklearn.externals import joblib

import spacy
from spacy.en import English

from features import get_questions_matrix_sum, get_images_matrix, get_answers_matrix, get_questions_tensor_timeseries
from utils import grouper, selectFrequentAnswers

def main():

	parser = argparse.ArgumentParser()
	parser.add_argument('-num_hidden_units_mlp', type=int, default=1024)
	parser.add_argument('-num_hidden_units_lstm', type=int, default=512)
	parser.add_argument('-num_hidden_layers_mlp', type=int, default=3)
	parser.add_argument('-num_hidden_layers_lstm', type=int, default=1)
	parser.add_argument('-dropout', type=float, default=0.5)
	parser.add_argument('-activation', type=str, default='tanh')
	parser.add_argument('-language_only', type=bool, default= False)
	parser.add_argument('-attention', type=bool, default= True)
	parser.add_argument('-num_epochs', type=int, default=100)
	parser.add_argument('-model_save_interval', type=int, default=5)
	parser.add_argument('-batch_size', type=int, default=128)
	args = parser.parse_args()

	questions_train = open('../data/preprocessed/questions_train2014.txt', 'r').read().decode('utf8').splitlines()
	questions_lengths_train = open('../data/preprocessed/questions_lengths_train2014.txt', 'r').read().decode('utf8').splitlines()
	answers_train = open('../data/preprocessed/answers_train2014_modal.txt', 'r').read().decode('utf8').splitlines()
	images_train = open('../data/preprocessed/images_train2014.txt', 'r').read().decode('utf8').splitlines()
	vgg_model_path = '../features/coco/vgg_feats.mat'

	img_dim = 4096
	word_vec_dim = 300
	max_len = 30
	nb_classes = 1000

	max_answers = nb_classes
	questions_train, answers_train, images_train = selectFrequentAnswers(questions_train,answers_train,images_train, max_answers)
	
	print 'Loaded questions, sorting by length...'
	questions_lengths_train, questions_train, answers_train, images_train = (list(t) for t in zip(*sorted(zip(questions_lengths_train, questions_train, answers_train, images_train))))

	#encode the remaining answers
	labelencoder = preprocessing.LabelEncoder()
	labelencoder.fit(answers_train)
	nb_classes = len(list(labelencoder.classes_))
	joblib.dump(labelencoder,'../models/labelencoder.pkl')

	nlp = spacy.load('en', vectors='en_glove_cc_300_1m_vectors') 
	print 'loaded Glove Vectors...'

	features_struct = scipy.io.loadmat(vgg_model_path)
	VGGfeatures = features_struct['feats']
	print 'loaded vgg features'
	image_ids = open('../features/coco_vgg_IDMap.txt').read().splitlines()
	id_map = {}
	for ids in image_ids:
		id_split = ids.split()
		id_map[id_split[0]] = int(id_split[1])

	language_model = Sequential()
	if args.num_hidden_layers_lstm == 1:
		language_model.add(LSTM(output_dim = args.num_hidden_units_lstm, return_sequences=False, input_shape=(max_len, word_vec_dim)))
	else:
		language_model.add(LSTM(output_dim = args.num_hidden_units_lstm, return_sequences=True, input_shape=(max_len, word_vec_dim)))
		for i in xrange(args.num_hidden_layers_lstm-2):
			language_model.add(LSTM(output_dim = args.num_hidden_units_lstm, return_sequences=True))
		language_model.add(LSTM(output_dim = args.num_hidden_units_lstm, return_sequences=False))

	attention_model = Sequential()
	attention_model.add(Dense(args.num_hidden_units_mlp, input_dim=word_vec_dim, init='uniform'))
	attention_model.add(Activation(args.activation))
	attention_model.add(Dense(img_dim, init='uniform'))
	# Can try softmax or tanh depending on how we interpret it
	attention_model.add(Activation('softmax'))
	
	attention_merge_model = Sequential()
	attention_merge_model.add(Reshape(input_shape = (img_dim,), dims=(img_dim,)))

	model = Graph()
	model.add_input(name='img_vec', input_shape=(img_dim,))
	model.add_input(name='word_vec', input_shape=(max_len, word_vec_dim))
	model.add_node(language_model, name='ques_vec', input='word_vec')
	model.add_node(attention_model, name='comp_att', input='ques_vec')
	model.add_node(attention_merge_model, name='att_merger', inputs=['img_vec', 'comp_att'], merge_mode='mul')
	model.add_node(Dense(args.num_hidden_units_mlp, init='uniform', activation=args.activation), name='ly0', inputs=['img_vec', 'att_merger', 'ques_vec'], merge_mode='concat')

	if args.dropout>0:
		model.add_node(Dropout(args.dropout), name='dropout0', input='ly0')

	for i in range(1,args.num_hidden_layers_mlp):
		model.add_node(Dense(args.num_hidden_units_mlp, init='uniform', activation=args.activation), name='ly'+str(i), input='dropout'+str(i-1))
		if args.dropout>0:
			model.add_node(Dropout(args.dropout), name='dropout'+str(i), input='ly'+str(i))
	model.add_node(Dense(nb_classes, init='uniform', activation='softmax'), name='final', input='dropout'+str(args.num_hidden_layers_mlp-1))
	model.add_output(name='output', input='final')

	json_string = model.to_json()
	model_file_name = '../models/lstm_att_num_hidden_units_lstm_' + str(args.num_hidden_units_lstm) + \
						'_num_hidden_units_mlp_' + str(args.num_hidden_units_mlp) + '_num_hidden_layers_mlp_' + \
						str(args.num_hidden_layers_mlp) + '_num_hidden_layers_lstm_' + str(args.num_hidden_layers_lstm)	
	open(model_file_name  + '.json', 'w').write(json_string)

	print 'Compiling model...'
	model.compile(optimizer='rmsprop', loss={'output': 'categorical_crossentropy'})
	print 'Compilation done...'
		
	print 'Training started...'
	for k in xrange(args.num_epochs):
		#shuffle the data points before going through them
		index_shuf = range(len(questions_train))
		shuffle(index_shuf)
		questions_train = [questions_train[i] for i in index_shuf]
		answers_train = [answers_train[i] for i in index_shuf]
		images_train = [images_train[i] for i in index_shuf]
		progbar = generic_utils.Progbar(len(questions_train))
		for qu_batch,an_batch,im_batch in zip(grouper(questions_train, args.batch_size, fillvalue=questions_train[-1]), 
											grouper(answers_train, args.batch_size, fillvalue=answers_train[-1]), 
											grouper(images_train, args.batch_size, fillvalue=images_train[-1])):
			timesteps = len(nlp(qu_batch[-1]))
			pad_vec = np.zeros((max_len - timesteps, word_vec_dim))
			X_q_batch =  get_questions_tensor_timeseries(qu_batch, nlp, timesteps)
			questions_tensor = np.zeros((args.batch_size, max_len, word_vec_dim))
			for m in range(0, len(X_q_batch)):
				questions_tensor[m] = np.append(X_q_batch[m], pad_vec, axis = 0)
			X_q_batch = np.array(questions_tensor)
			print (X_q_batch).shape
			if args.language_only:
				X_batch = X_q_batch
			else:
				X_i_batch = get_images_matrix(im_batch, id_map, VGGfeatures)
			Y_batch = get_answers_matrix(an_batch, labelencoder)
			print X_i_batch.shape
			print X_q_batch.shape
			loss = model.train_on_batch({'img_vec': X_i_batch, 'word_vec': X_q_batch, 'output': Y_batch})
			progbar.add(args.batch_size, values=[("train loss", loss)])

			model.test_on_batch(data, accuracy=True, sample_weight={})
		#print type(loss)
		if k%args.model_save_interval == 0:
			model.save_weights(model_file_name + '_epoch_{:02d}.hdf5'.format(k))

	model.save_weights(model_file_name + '_epoch_{:02d}.hdf5'.format(k))

if __name__ == "__main__":
	main()
